@@ -1,12 +1,12 @@
-# Copyright 2014-2015 Canonical Limited.
+# Copyright 2014-2017 Canonical Limited.
 #
-# This file is part of charm-helpers.
+# This file is part of charms.reactive
 #
-# charm-helpers is free software: you can redistribute it and/or modify
+# charms.reactive is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License version 3 as
 # published by the Free Software Foundation.
 #
-# charm-helpers is distributed in the hope that it will be useful,
+# charms.reactive is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU Lesser General Public License for more details.
@@ -15,21 +15,209 @@
 # along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import sys
+import importlib
 from inspect import isclass
 
-from six import with_metaclass
-
 from charmhelpers.core import hookenv
 from charmhelpers.core import unitdata
 from charmhelpers.cli import cmdline
-from charms.reactive.bus import get_state
-from charms.reactive.bus import set_state
-from charms.reactive.bus import remove_state
-from charms.reactive.bus import _load_module
-from charms.reactive.bus import StateList
+from charms.reactive.flags import get_flags
+from charms.reactive.flags import _get_flag_value
+from charms.reactive.flags import set_flag
+from charms.reactive.flags import clear_flag
+from charms.reactive.flags import StateList
+from charms.reactive.bus import _append_path
+
+try:
+    from importlib.metadata import entry_points
+except ImportError:
+    from pkg_resources import iter_entry_points
+
+    def entry_points():
+        group = 'charms.reactive.relation_factory'
+        return {group: list(iter_entry_points(group))}
+
+__all__ = [
+    'endpoint_from_name',
+    'endpoint_from_flag',
+    'relation_from_flag',  # DEPRECATED
+    'scopes',  # DEPRECATED
+    'RelationBase',  # DEPRECATED
+    'relation_from_state',  # DEPRECATED
+]
 
 
-ALL = '__ALL_SERVICES__'
+# arbitrary obj instances to use as defaults instead of None
+ALL = object()
+TOGGLE = object()
+
+
+def endpoint_from_name(endpoint_name):
+    """The object used for interacting with the named relations, or None.
+    """
+    if endpoint_name is None:
+        return None
+    factory = relation_factory(endpoint_name)
+    if factory:
+        return factory.from_name(endpoint_name)
+
+
+def relation_from_name(relation_name):
+    """
+    .. deprecated:: 0.6.0
+       Alias for :func:`endpoint_from_name`
+    """
+    return endpoint_from_name(relation_name)
+
+
+def endpoint_from_flag(flag):
+    """The object used for interacting with relations tied to a flag, or None.
+    """
+    relation_name = None
+    value = _get_flag_value(flag)
+    if isinstance(value, dict) and 'relation' in value:
+        # old-style RelationBase
+        relation_name = value['relation']
+    elif flag.startswith('endpoint.'):
+        # new-style Endpoint
+        relation_name = flag.split('.')[1]
+    elif '.' in flag:
+        # might be an unprefixed new-style Endpoint
+        relation_name = flag.split('.')[0]
+        if relation_name not in hookenv.relation_types():
+            return None
+    if relation_name:
+        factory = relation_factory(relation_name)
+        if factory:
+            return factory.from_flag(flag)
+    return None
+
+
+def relation_from_flag(flag):
+    """
+    .. deprecated:: 0.6.0
+       Alias for :func:`endpoint_from_flag`
+    """
+    return endpoint_from_flag(flag)
+
+
+def relation_from_state(state):
+    """
+    .. deprecated:: 0.5.0
+       Alias for :func:`endpoint_from_flag`
+    """
+    return endpoint_from_flag(state)
+
+
+class RelationFactory(object):
+    """Produce objects for interacting with a relation.
+
+    Interfaces choose which RelationFactory is used for their relations
+    by adding a RelationFactory subclass to
+    ``$CHARM_DIR/hooks/relations/{interface}/{provides,requires,peer}.py``.
+    This is normally a RelationBase subclass.
+    """
+    _factories = []
+
+    @classmethod
+    def discover(cls):
+        for ep in entry_points().get('charms.reactive.relation_factory', []):
+            factory = ep.load()
+            factory.load()
+            RelationFactory._factories.append(factory)
+
+    @classmethod
+    def get_factory(cls, relation_name):
+        for factory in RelationFactory._factories:
+            if factory.from_name(relation_name):
+                return factory
+        return None
+
+    @classmethod
+    def from_name(cls, relation_name):
+        raise NotImplementedError()
+
+    @classmethod
+    def from_flag(cls, state):
+        raise NotImplementedError()
+
+
+def relation_factory(relation_name):
+    """Get the RelationFactory for the given relation name.
+
+    Looks for a RelationFactory in the first file matching:
+    ``$CHARM_DIR/hooks/relations/{interface}/{provides,requires,peer}.py``
+    """
+    factory = RelationFactory.get_factory(relation_name)
+    if factory:
+        return factory
+    role, interface = hookenv.relation_to_role_and_interface(relation_name)
+    if not (role and interface):
+        hookenv.log('Unable to determine role and interface for relation '
+                    '{}'.format(relation_name), hookenv.ERROR)
+        return None
+    return _find_relation_factory(_relation_module(role, interface))
+
+
+def _relation_module(role, interface):
+    """
+    Return module for relation based on its role and interface, or None.
+
+    Prefers new location (reactive/relations) over old (hooks/relations).
+    """
+    _append_path(hookenv.charm_dir())
+    _append_path(os.path.join(hookenv.charm_dir(), 'hooks'))
+    base_module = 'relations.{}.{}'.format(interface, role)
+    for module in ('reactive.{}'.format(base_module), base_module):
+        if module in sys.modules:
+            break
+        try:
+            importlib.import_module(module)
+            break
+        except ImportError:
+            continue
+    else:
+        hookenv.log('Unable to find implementation for relation: '
+                    '{} of {}'.format(role, interface), hookenv.ERROR)
+        return None
+    return sys.modules[module]
+
+
+def _find_relation_factory(module):
+    """
+    Attempt to find a RelationFactory subclass in the module.
+
+    Note: RelationFactory and RelationBase are ignored so they may
+    be imported to be used as base classes without fear.
+    """
+    if not module:
+        return None
+
+    # All the RelationFactory subclasses
+    candidates = [o for o in (getattr(module, attr) for attr in dir(module))
+                  if (o is not RelationFactory and
+                      o is not RelationBase and
+                      isclass(o) and
+                      issubclass(o, RelationFactory))]
+
+    # Filter out any factories that are superclasses of another factory
+    # (none of the other factories subclass it). This usually makes
+    # the explict check for RelationBase and RelationFactory unnecessary.
+    candidates = [c1 for c1 in candidates
+                  if not any(issubclass(c2, c1) for c2 in candidates
+                             if c1 is not c2)]
+
+    if not candidates:
+        hookenv.log('No RelationFactory found in {}'.format(module.__name__),
+                    hookenv.WARNING)
+        return None
+
+    if len(candidates) > 1:
+        raise RuntimeError('Too many RelationFactory found in {}'
+                           ''.format(module.__name__))
+
+    return candidates[0]
 
 
 class scopes(object):
@@ -89,9 +277,9 @@ class AutoAccessors(type):
         return __accessor
 
 
-class RelationBase(with_metaclass(AutoAccessors, object)):
+class RelationBase(RelationFactory, metaclass=AutoAccessors):
     """
-    The base class for all relation implementations.
+    A base class for relation implementations.
     """
     _cache = {}
 
@@ -148,6 +336,16 @@ class RelationBase(with_metaclass(AutoAccessors, object)):
     recommended that this be used only with :attr:`scopes.GLOBAL` scope.
     """
 
+    @classmethod
+    def _startup(cls):
+        # update data to be backwards compatible after fix for issue 28
+        _migrate_conversations()
+
+        if hookenv.hook_name().endswith('-relation-departed'):
+            def depart_conv():
+                cls(hookenv.relation_type()).conversation().depart()
+            hookenv.atexit(depart_conv)
+
     def __init__(self, relation_name, conversations=None):
         self._relation_name = relation_name
         self._conversations = conversations or [Conversation.join(self.scope)]
@@ -162,10 +360,21 @@ class RelationBase(with_metaclass(AutoAccessors, object)):
     @classmethod
     def from_state(cls, state):
         """
-        Find relation implementation in the current charm, based on the
-        name of an active state.
+        .. deprecated:: 0.6.1
+           use :func:`endpoint_from_flag` instead
         """
-        value = get_state(state)
+        return cls.from_flag(state)
+
+    @classmethod
+    def from_flag(cls, flag):
+        """
+        Find relation implementation in the current charm, based on the
+        name of an active flag.
+
+        You should not use this method directly.
+        Use :func:`endpoint_from_flag` instead.
+        """
+        value = _get_flag_value(flag)
         if value is None:
             return None
         relation_name = value['relation']
@@ -176,7 +385,7 @@ class RelationBase(with_metaclass(AutoAccessors, object)):
     def from_name(cls, relation_name, conversations=None):
         """
         Find relation implementation in the current charm, based on the
-        ID of the relation.
+        name of the relation.
 
         :return: A Relation instance, or None
         """
@@ -197,17 +406,11 @@ class RelationBase(with_metaclass(AutoAccessors, object)):
     def _find_impl(cls, role, interface):
         """
         Find relation implementation based on its role and interface.
-
-        Looks for the first file matching:
-        ``$CHARM_DIR/hooks/relations/{interface}/{provides,requires,peer}.py``
         """
-        hooks_dir = os.path.join(hookenv.charm_dir(), 'hooks')
-        try:
-            filepath = os.path.join(hooks_dir, 'relations', interface, role + '.py')
-            module = _load_module(filepath)
-            return cls._find_subclass(module)
-        except ImportError:
+        module = _relation_module(role, interface)
+        if not module:
             return None
+        return cls._find_subclass(module)
 
     @classmethod
     def _find_subclass(cls, module):
@@ -220,7 +423,8 @@ class RelationBase(with_metaclass(AutoAccessors, object)):
         """
         for attr in dir(module):
             candidate = getattr(module, attr)
-            if isclass(candidate) and issubclass(candidate, cls) and candidate is not cls:
+            if (isclass(candidate) and issubclass(candidate, cls) and
+                    candidate is not RelationBase):
                 return candidate
         return None
 
@@ -283,13 +487,37 @@ class RelationBase(with_metaclass(AutoAccessors, object)):
         """
         self.conversation(scope).remove_state(state)
 
+    def is_state(self, state, scope=None):
+        """
+        Test the state for the :class:`Conversation` with the given scope.
+
+        In Python, this is equivalent to::
+
+            relation.conversation(scope).is_state(state)
+
+        See :meth:`conversation` and :meth:`Conversation.is_state`.
+        """
+        return self.conversation(scope).is_state(state)
+
+    def toggle_state(self, state, active=TOGGLE, scope=None):
+        """
+        Toggle the state for the :class:`Conversation` with the given scope.
+
+        In Python, this is equivalent to::
+
+            relation.conversation(scope).toggle_state(state, active)
+
+        See :meth:`conversation` and :meth:`Conversation.toggle_state`.
+        """
+        self.conversation(scope).toggle_state(state, active)
+
     def set_remote(self, key=None, value=None, data=None, scope=None, **kwdata):
         """
         Set data for the remote end(s) of the :class:`Conversation` with the given scope.
 
         In Python, this is equivalent to::
 
-            relation.conversation(scope).set_remote(data, scope, **kwdata)
+            relation.conversation(scope).set_remote(key, value, data, scope, **kwdata)
 
         See :meth:`conversation` and :meth:`Conversation.set_remote`.
         """
@@ -353,31 +581,39 @@ class Conversation(object):
     Conversations use the idea of :class:`scope` to determine how units and
     services are grouped together.
     """
-    def __init__(self, relation_name=None, units=None, scope=None):
-        self.relation_name = relation_name or hookenv.relation_type()
-        self.units = set(units or [hookenv.remote_unit()])
-        self.scope = scope or hookenv.remote_unit()
+    def __init__(self, namespace, units, scope):
+        self.namespace = namespace
+        self.units = set(units)
+        self.scope = scope
+
+    @classmethod
+    def _key(cls, namespace, scope):
+        return 'reactive.conversations.%s.%s' % (namespace, scope)
 
     @property
     def key(self):
         """
         The key under which this conversation will be stored.
         """
-        return 'reactive.conversations.%s.%s' % (self.relation_name, self.scope)
+        return self._key(self.namespace, self.scope)
+
+    @property
+    def relation_name(self):
+        return self.namespace.split(':')[0]
 
     @property
-    @hookenv.cached
     def relation_ids(self):
         """
         The set of IDs of the specific relation instances that this conversation
         is communicating with.
         """
-        relation_ids = []
-        services = set(unit.split('/')[0] for unit in self.units)
-        for relation_id in hookenv.relation_ids(self.relation_name):
-            if hookenv.remote_service_name(relation_id) in services:
-                relation_ids.append(relation_id)
-        return relation_ids
+        if self.scope == scopes.GLOBAL:
+            # the namespace is the relation name and this conv speaks for all
+            # connected instances of that relation
+            return hookenv.relation_ids(self.namespace)
+        else:
+            # the namespace is the relation ID
+            return [self.namespace]
 
     @classmethod
     def join(cls, scope):
@@ -391,14 +627,20 @@ class Conversation(object):
         :meth:`~charmhelpers.core.unitdata.Storage.flush` be called.
         """
         relation_name = hookenv.relation_type()
+        relation_id = hookenv.relation_id()
         unit = hookenv.remote_unit()
         service = hookenv.remote_service_name()
         if scope is scopes.UNIT:
             scope = unit
+            namespace = relation_id
         elif scope is scopes.SERVICE:
             scope = service
-        key = 'reactive.conversations.%s.%s' % (relation_name, scope)
-        conversation = cls.deserialize(unitdata.kv().get(key, {'scope': scope}))
+            namespace = relation_id
+        else:
+            namespace = relation_name
+        key = cls._key(namespace, scope)
+        data = unitdata.kv().get(key, {'namespace': namespace, 'scope': scope, 'units': []})
+        conversation = cls.deserialize(data)
         conversation.units.add(unit)
         unitdata.kv().set(key, cls.serialize(conversation))
         return conversation
@@ -407,9 +649,6 @@ class Conversation(object):
         """
         Remove the current remote unit, for the active hook context, from
         this conversation.  This should be called from a `-departed` hook.
-
-        TODO: Need to figure out a way to have this called implicitly, to
-        ensure cleaning up of conversations that are no longer needed.
         """
         unit = hookenv.remote_unit()
         self.units.remove(unit)
@@ -431,8 +670,8 @@ class Conversation(object):
         Serialize a conversation instance for storage.
         """
         return {
-            'relation_name': conversation.relation_name,
-            'units': list(conversation.units),
+            'namespace': conversation.namespace,
+            'units': sorted(conversation.units),
             'scope': conversation.scope,
         }
 
@@ -466,12 +705,13 @@ class Conversation(object):
         :meth:`~charmhelpers.core.unitdata.Storage.flush` be called.
         """
         state = state.format(relation_name=self.relation_name)
-        value = get_state(state, {
+        value = _get_flag_value(state, {
             'relation': self.relation_name,
             'conversations': [],
         })
-        value['conversations'].append(self.key)
-        set_state(state, value)
+        if self.key not in value['conversations']:
+            value['conversations'].append(self.key)
+        set_flag(state, value)
 
     def remove_state(self, state):
         """
@@ -489,15 +729,48 @@ class Conversation(object):
         conversations are in this the state, will deactivate it.
         """
         state = state.format(relation_name=self.relation_name)
-        value = get_state(state)
+        value = _get_flag_value(state)
         if not value:
             return
         if self.key in value['conversations']:
             value['conversations'].remove(self.key)
         if value['conversations']:
-            set_state(state, value)
+            set_flag(state, value)
         else:
-            remove_state(state)
+            clear_flag(state)
+
+    def is_state(self, state):
+        """
+        Test if this conversation is in the given state.
+        """
+        state = state.format(relation_name=self.relation_name)
+        value = _get_flag_value(state)
+        if not value:
+            return False
+        return self.key in value['conversations']
+
+    def toggle_state(self, state, active=TOGGLE):
+        """
+        Toggle the given state for this conversation.
+
+        The state will be set ``active`` is ``True``, otherwise the state will be removed.
+
+        If ``active`` is not given, it will default to the inverse of the current state
+        (i.e., ``False`` if the state is currently set, ``True`` if it is not; essentially
+        toggling the state).
+
+        For example::
+
+            conv.toggle_state('{relation_name}.foo', value=='foo')
+
+        This will set the state if ``value`` is equal to ``foo``.
+        """
+        if active is TOGGLE:
+            active = not self.is_state(state)
+        if active:
+            self.set_state(state)
+        else:
+            self.remove_state(state)
 
     def set_remote(self, key=None, value=None, data=None, **kwdata):
         """
@@ -513,9 +786,9 @@ class Conversation(object):
         the data will be set for all of those services.
 
         :param str key: The name of a field to set.
-        :param value: A value to set.
+        :param value: A value to set. This value must be json serializable.
         :param dict data: A mapping of keys to values.
-        :param \*\*kwdata: A mapping of keys to values, as keyword arguments.
+        :param **kwdata: A mapping of keys to values, as keyword arguments.
         """
         if data is None:
             data = {}
@@ -537,8 +810,16 @@ class Conversation(object):
         converging to identical data.  Thus, this method returns the first
         value that it finds set by any of its units.
         """
+        cur_rid = hookenv.relation_id()
+        departing = hookenv.hook_name().endswith('-relation-departed')
         for relation_id in self.relation_ids:
-            for unit in hookenv.related_units(relation_id):
+            units = hookenv.related_units(relation_id)
+            if departing and cur_rid == relation_id:
+                # Work around the fact that Juju 2.0 doesn't include the
+                # departing unit in relation-list during the -departed hook,
+                # by adding it back in ourselves.
+                units.append(hookenv.remote_unit())
+            for unit in units:
                 if unit not in self.units:
                     continue
                 value = hookenv.relation_get(key, unit, relation_id)
@@ -565,9 +846,9 @@ class Conversation(object):
         :meth:`~charmhelpers.core.unitdata.Storage.flush` be called.
 
         :param str key: The name of a field to set.
-        :param value: A value to set.
+        :param value: A value to set. This value must be json serializable.
         :param dict data: A mapping of keys to values.
-        :param \*\*kwdata: A mapping of keys to values, as keyword arguments.
+        :param **kwdata: A mapping of keys to values, as keyword arguments.
         """
         if data is None:
             data = {}
@@ -586,21 +867,70 @@ class Conversation(object):
         return unitdata.kv().get(key, default)
 
 
+def _migrate_conversations():  # noqa
+    """
+    Due to issue #28 (https://github.com/juju-solutions/charms.reactive/issues/28),
+    conversations needed to be updated to be namespaced per relation ID for SERVICE
+    and UNIT scope.  To ensure backwards compatibility, this updates all convs in
+    the old format to the new.
+
+    TODO: Remove in 2.0.0
+    """
+    for key, data in unitdata.kv().getrange('reactive.conversations.').items():
+        if 'local-data' in key:
+            continue
+        if 'namespace' in data:
+            continue
+        relation_name = data.pop('relation_name')
+        if data['scope'] == scopes.GLOBAL:
+            data['namespace'] = relation_name
+            unitdata.kv().set(key, data)
+        else:
+            # split the conv based on the relation ID
+            new_keys = []
+            for rel_id in hookenv.relation_ids(relation_name):
+                new_key = Conversation._key(rel_id, data['scope'])
+                new_units = set(hookenv.related_units(rel_id)) & set(data['units'])
+                if new_units:
+                    unitdata.kv().set(new_key, {
+                        'namespace': rel_id,
+                        'scope': data['scope'],
+                        'units': sorted(new_units),
+                    })
+                    new_keys.append(new_key)
+            unitdata.kv().unset(key)
+            # update the states pointing to the old conv key to point to the
+            # (potentially multiple) new key(s)
+            for flag in get_flags():
+                value = _get_flag_value(flag)
+                if not value:
+                    continue
+                if key not in value['conversations']:
+                    continue
+                value['conversations'].remove(key)
+                value['conversations'].extend(new_keys)
+                set_flag(flag, value)
+
+
 @cmdline.subcommand()
-def relation_call(method, relation_name=None, state=None, *args):
+def relation_call(method, relation_name=None, flag=None, state=None, *args):
     """Invoke a method on the class implementing a relation via the CLI"""
     if relation_name:
-        relation = RelationBase.from_name(relation_name)
+        relation = relation_from_name(relation_name)
         if relation is None:
             raise ValueError('Relation not found: %s' % relation_name)
-    elif state:
-        relation = RelationBase.from_state(state)
+    elif flag or state:
+        relation = relation_from_flag(flag or state)
         if relation is None:
-            raise ValueError('Relation not found: %s' % state)
+            raise ValueError('Relation not found: %s' % (flag or state))
     else:
-        raise ValueError('Must specify either relation_name or state')
+        raise ValueError('Must specify either relation_name or flag')
     result = getattr(relation, method)(*args)
-    if method == 'conversations':
+    if isinstance(relation, RelationBase) and method == 'conversations':
         # special case for conversations to make them work from CLI
         result = [c.scope for c in result]
     return result
+
+
+hookenv.atstart(RelationBase._startup)
+RelationFactory.discover()
